import io

from pykeepass import PyKeePass, create_database
from pykeepass.entry import Entry
from pykeepass.group import Group


def create_kdb(password: str = "test123") -> PyKeePass:
    file = io.BytesIO()
    return create_database(filename=file, password=password)


def create_kdb_group(
    kdb: PyKeePass,
    group_name: str,
    parent_group: Group = None,
    icon: str = "1",
) -> Group:
    if not parent_group:
        parent_group = kdb.root_group
    return kdb.add_group(
        destination_group=parent_group, group_name=group_name, icon=icon
    )


def create_kdb_entry(
    kdb: PyKeePass, parent_group: Group = None, **kwargs
) -> Entry:
    if not parent_group:
        parent_group = kdb.root_group
    return kdb.add_entry(destination_group=parent_group, **kwargs)



import io
from unittest import TestCase

from pykeepass import PyKeePass

from kdb_vault_tools import Processor
from kdb_vault_tools.processors import KDBProcessor, VaultProcessor

from .utils import create_kdb, create_kdb_entry, create_kdb_group


class TestKDBVaultFeature(TestCase):
    def setUp(self):
        self.kdb_password = "test123"
        self.vault_config = {"token": "myroot", "url": "http://127.0.0.1:1234"}
        file_like = io.BytesIO()

        self.kdb = create_kdb(self.kdb_password)
        entry_data = {
            "icon": "1",
            "password": "root",
            "url": "https://127.0.0.1/",
            "username": "tester",
            "notes": "",
        }
        group = create_kdb_group(self.kdb, "test")
        group = create_kdb_group(self.kdb, "sub", parent_group=group)
        create_kdb_entry(
            self.kdb, parent_group=group, title="test_entry", **entry_data
        )
        self.kdb.save(file_like)
        file_like.seek(0)
        self.kdb_config = {
            "filename": file_like,
            "password": self.kdb_password,
        }

    def test_create(self):
        processor = Processor({"conf": 1}, {"conf": 2})
        self.assertIsInstance(processor._vault_processor, VaultProcessor)
        self.assertIsInstance(processor._kdb_processor, KDBProcessor)
        self.assertEqual(processor._kdb_settings["conf"], 2)
        self.assertEqual(processor._vault_settings["conf"], 1)

        self.assertTrue(
            processor.kdb_groups
            == processor._vault_processor.kdb_groups
            == processor._kdb_processor.kdb_groups
        )
        self.assertTrue(
            processor.kdb_entries
            == processor._vault_processor.kdb_entries
            == processor._kdb_processor.kdb_entries
        )

    def test_processor(self):
        processor = Processor(
            vault_settings=self.vault_config, kdb_settings=self.kdb_config
        )
        processor.sync_from_kdb()
        processor.sync_from_vault()
        processor.write_vault()
        file_like = io.BytesIO()
        processor._kdb_settings["filename"] = file_like
        processor.write_kdb()
        file_like.seek(0)
        kdb = PyKeePass(filename=file_like, password=self.kdb_password)
        entry = kdb.find_entries_by_title(title="test_entry", first=True)
        self.assertEqual(entry.path, ["test", "sub", "test_entry"])



from unittest import TestCase

from hvac.exceptions import InvalidPath

from kdb_vault_tools.datatypes.basic import BaseEntry, BaseGroup
from kdb_vault_tools.exceptions import VaultConnectionError
from kdb_vault_tools.processors import VaultProcessor


class TestVaultFeature(TestCase):
    def setUp(self):
        self.vault_config = {"token": "myroot", "url": "http://127.0.0.1:1234"}
        self.processor = VaultProcessor()
        self.processor.vault_connect(**self.vault_config)

    def test_connection(self):
        try:
            self.processor.vault_connect()
            raise Exception("Failed")
        except VaultConnectionError as err:
            self.assertEqual(str(err), "Vault not Configured")

        self.processor.vault_connect(**self.vault_config)
        self.assertTrue(self.processor.vault.is_authenticated())

    def _test_write_secret(self, entry, base_path="/"):
        self.processor.write_secret(entry=entry, base_path=base_path)
        secret_full_path = f"{base_path}{entry.path}"
        resp = self.processor.vault.secrets.kv.v2.read_secret_version(
            path=secret_full_path
        )
        data = resp["data"]["data"]
        expected_data = entry.as_dict()
        expected_data.pop("path")
        expected_data.pop("title")
        self.assertEqual(
            data,
            expected_data,
        )

    def test_write_secret(self):
        entry = BaseEntry(
            title="root secret",
            password="root",
            icon="1",
            path="root_secret",
            url="",
            meta={"foo": "bar"},
            username="odin",
        )
        self._test_write_secret(entry)
        entry.title = "root2"
        self._test_write_secret(entry, "/foo/bar/baz/")
        entry.title = "root3"
        self._test_write_secret(entry, "/foo/bar/baz/")
        entry.title = "root4"
        self._test_write_secret(entry, "/foo/")
        try:
            self._test_write_secret(entry, "/foo/   ")
        except InvalidPath:
            pass

    def _test_write_group(self, group, base_path="/"):
        self.processor.write_group(group, base_path=base_path)
        group_meta_full_path = (
            f"{base_path}{group.path}/{self.processor.group_meta_label}"
        )
        resp = self.processor.vault.secrets.kv.v2.read_secret_version(
            path=group_meta_full_path
        )
        data = resp["data"]["data"]
        expected_data = group.as_dict()
        expected_data.pop("path")
        expected_data.pop("title")
        self.assertEqual(
            data,
            expected_data,
        )

    def test_write_group(self):
        entry = BaseGroup(
            title="root_group",
            icon="42",
            path="root_group",
        )
        self._test_write_group(entry)
        self._test_write_group(entry, "/main_group/")
        self._test_write_group(entry, "/main_group/sub/")
        self._test_write_group(entry, "/main_group/sub/sub/")

    def test_read_secret(self):
        entry = BaseEntry(
            title="test_read",
            password="root",
            icon="1",
            path="test_read",
            url="",
            meta={"foo": "bar"},
            username="odin",
        )
        self.processor.write_secret(entry)
        self.processor.read_secret(f"/{entry.path}")
        self.assertEqual(self.processor.kdb_entries[""][0], entry)

    def test_read_secrets(self):
        base_path = "test/recursive/read/"

        group_entry = BaseGroup(
            title="root_group",
            icon="42",
            path=f"{base_path}root_group",
        )

        secret_entry = BaseEntry(
            title="test_entry",
            password="root",
            icon="1",
            path=f"{base_path}root_group/test_entry",
            url="",
            meta={"foo": "bar"},
            username="odin",
        )

        self.processor.write_group(group=group_entry)
        self.processor.write_secret(entry=secret_entry)
        self.processor.read_secrets(path=f"/{base_path}", recursive=True)

        self.assertEqual(
            self.processor.kdb_groups[group_entry.path], group_entry
        )
        self.assertEqual(
            self.processor.kdb_entries[group_entry.path][0], secret_entry
        )

    def test_build_vault(self):
        base_path = "test/recursive/build/"

        group_entry = BaseGroup(
            title="root_group",
            icon="42",
            path=f"{base_path}root_group",
        )

        secret_entry = BaseEntry(
            title="test_entry",
            password="root",
            icon="1",
            path=f"{base_path}root_group/test_entry",
            url="",
            meta={"foo": "bar"},
            username="odin",
        )
        self.processor.kdb_entries[group_entry.path] = [
            secret_entry,
        ]
        self.processor.kdb_groups[group_entry.path] = group_entry

        self.processor.build_vault()

        # TODO DRY
        resp = self.processor.vault.secrets.kv.v2.read_secret_version(
            path=f"{group_entry.path}/{self.processor.group_meta_label}"
        )
        data = resp["data"]["data"]
        expected_data = group_entry.as_dict()
        expected_data.pop("path")
        expected_data.pop("title")
        self.assertEqual(
            data,
            expected_data,
        )

        resp = self.processor.vault.secrets.kv.v2.read_secret_version(
            path=f"{group_entry.path}/{secret_entry.title}"
        )
        data = resp["data"]["data"]
        expected_data = secret_entry.as_dict()
        expected_data.pop("path")
        expected_data.pop("title")
        self.assertEqual(
            data,
            expected_data,
        )



import dataclasses
import json
import os
from typing import Optional

from pykeepass.entry import Entry
from pykeepass.group import Group


@dataclasses.dataclass
class BaseGroup:
    title: str
    path: str
    icon: Optional[str]

    @classmethod
    def from_kdb_entry(cls, entry: Group):
        return cls(
            title=entry.name, path="/".join(entry.path), icon=entry.icon
        )

    def as_dict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass
class BaseEntry:
    title: str
    path: str
    password: str
    meta: Optional[dict]
    icon: Optional[str]
    username: str = ""
    url: str = ""

    @classmethod
    def from_kdb_entry(cls, entry: Entry):
        return cls(
            password=entry.password,
            url=entry.url,
            username=entry.username,
            # notes=entry.notes,
            path=os.path.join(*entry.path),
            title=entry.title,
            icon=entry.icon,
            meta=cls.parse_notes(entry.notes),
        )

    @staticmethod
    def parse_notes(notes: str) -> dict:
        # TODO use template lang
        meta = {}

        if not notes:
            return meta

        for line in notes.split("\n"):
            if ":" in line:
                key, *content = line.split(":", maxsplit=2)
            else:
                key, *content = line.split(" ", maxsplit=2)

            meta[key] = "".join(content).strip()

        return meta

    @staticmethod
    def dump_notes(meta: dict) -> str:
        notes = []

        for key, content in meta.items():
            notes.append(f"{key}: {content}")

        return "\n".join(notes)

    def dump_json(self, indent=None) -> str:
        return json.dumps(self.as_dict(), indent=indent)

    def as_dict(self) -> dict:
        return dataclasses.asdict(self)

    @property
    def group_path(self) -> str:
        return self.path.replace(self.title, "").rstrip("/")



import io
import os
from unittest import TestCase

from pykeepass import PyKeePass
from pykeepass.group import Group

from kdb_vault_tools.datatypes.basic import BaseEntry, BaseGroup
from kdb_vault_tools.exceptions import EntryBuildError
from kdb_vault_tools.processors import KDBProcessor

from .utils import create_kdb, create_kdb_entry, create_kdb_group


class TestKDBFeature(TestCase):
    def setUp(self):
        self.password = "123456"
        kdb = create_kdb(password=self.password)
        entry_data = {
            "icon": "1",
            "password": "root",
            "url": "https://127.0.0.1/",
            "username": "tester",
            "notes": "",
        }
        create_kdb_entry(kdb, title="root_entry", **entry_data)
        kdb_sub_grp = create_kdb_group(kdb, group_name="sub")
        create_kdb_entry(
            kdb, title="sub_entry", **entry_data, parent_group=kdb_sub_grp
        )
        kdb_sub_sub_grp = create_kdb_group(
            kdb, group_name="subsub", parent_group=kdb_sub_grp
        )
        create_kdb_entry(
            kdb,
            title="sub_sub_entry",
            **entry_data,
            parent_group=kdb_sub_sub_grp,
        )
        self.raw_kdb_full_path = "/tmp/__test_kdb.kdbx"
        kdb.save(filename=self.raw_kdb_full_path)
        self.kdb = kdb
        self.processor = KDBProcessor()

    def tearDown(self):
        os.remove(self.raw_kdb_full_path)

    def test_creation(self):
        self.processor.load_from_kdb(self.raw_kdb_full_path, self.password)
        self.assertEqual(
            list(self.processor.kdb_entries.keys()), ["", "sub", "sub/subsub"]
        )
        self.assertEqual(
            list(self.processor.kdb_groups.keys()), ["", "sub", "sub/subsub"]
        )

    def test_build_kdb(self):
        file = self.raw_kdb_full_path
        self.processor.build_kdb(filename=file, password="123")

        kdb = PyKeePass(filename=file, password="123")
        entries = kdb.entries
        groups = kdb.groups
        self.assertEqual(len(entries), 0)
        self.assertEqual(len(groups), 1)

        kdb_io = io.BytesIO()
        self.kdb.save(kdb_io)
        kdb_io.seek(0)
        self.processor.load_from_kdb(kdb_io, self.password)
        self.processor.build_kdb(filename=file, password="123")

        kdb = PyKeePass(filename=file, password="123")
        entries = kdb.entries
        groups = kdb.groups
        self.assertEqual(3, len(entries))
        self.assertEqual(3, len(groups))

        # TODO DRY
        self.processor.build_kdb(filename=kdb_io, password="123")

        kdb = PyKeePass(filename=file, password="123")
        entries = kdb.entries
        groups = kdb.groups
        self.assertEqual(3, len(entries))
        self.assertEqual(3, len(groups))

    def test_load_entry(self):
        kdb_entry = self.kdb.entries[0]
        self.processor._load_entry_from_kdb(kdb_entry)
        self.assertTrue("" in self.processor.kdb_entries)
        self.assertEqual(len(self.processor.kdb_entries[""]), 1)
        entry = self.processor.kdb_entries[""][0]
        self.assertIsInstance(entry, BaseEntry)
        self.assertEqual(
            entry,
            BaseEntry(
                title="root_entry",
                username="tester",
                meta={},
                icon="1",
                password="root",
                url="https://127.0.0.1/",
                path="root_entry",
            ),
        )

    def test_load_group_from_kdb(self):
        kdb_entry = self.kdb.groups[0]
        self.processor._load_group_from_kdb(kdb_entry, load_entries=False)
        self.assertTrue("" in self.processor.kdb_groups)
        group = self.processor.kdb_groups[""]
        self.assertIsInstance(group, BaseGroup)
        self.assertEqual(group, BaseGroup(title="Root", icon="48", path=""))
        self.assertEqual(0, len(self.processor.kdb_entries.keys()))
        self.processor._load_group_from_kdb(kdb_entry, load_entries=True)
        self.assertEqual(group, BaseGroup(title="Root", icon="48", path=""))
        self.assertEqual(1, len(self.processor.kdb_entries.keys()))

    def test_exception_handler(self):
        exc = Exception("TEST")
        res = self.processor.exception_handler(exception=exc)
        self.assertIsNone(res)

    def test_init_kdb_connection(self):
        res = self.processor.init_kdb_connection()
        self.assertIsNone(res)

    def test_build_from_blank(self):

        processor = KDBProcessor()
        processor._blank_kdb_file = self.raw_kdb_full_path
        processor._blank_kdb_file_password = self.password
        processor._build_from_blank("Foo", "test123")

        entries = processor.kdb.entries
        groups = processor.kdb.groups
        self.assertEqual(3, len(entries))
        self.assertEqual(3, len(groups))

    def test_kdb_group_get_or_create(self):
        self.processor.load_from_kdb(self.raw_kdb_full_path, self.password)

        kdb_group = self.processor._kdb_group_get_or_create("sub/subsub")
        self.assertIsInstance(kdb_group, Group)
        self.assertEqual(kdb_group.path, ["sub", "subsub"])
        self.assertEqual(kdb_group.name, "subsub")
        self.assertEqual(kdb_group.icon, "1")

        self.processor.kdb_groups["sub/subsub/created"] = BaseGroup(
            path="sub/subsub/created",
            title="created",
            icon="42",
        )

        kdb_group = self.processor._kdb_group_get_or_create(
            "sub/subsub/created"
        )
        self.assertIsInstance(kdb_group, Group)
        self.assertEqual(kdb_group.path, ["sub", "subsub", "created"])
        self.assertEqual(kdb_group.name, "created")
        self.assertEqual(kdb_group.icon, "42")

    def test_save_groups_to_kdb(self):
        self.processor.load_from_kdb(self.raw_kdb_full_path, self.password)
        self.processor.kdb_groups[
            "sub/Foo][ \ 1 *8!@#$%^&*()/ [ subsub/created"
        ] = BaseGroup(
            path="sub/su/created",
            title="crea[ ']ted",
            icon="42",
        )
        self.processor._save_groups_to_kdb()

        self.assertFalse("crea[ ']ted" in self.kdb.groups)
        self.assertFalse("created" in self.kdb.groups)

    def test_exceptions_group_get_or_create(self):
        self.processor.load_from_kdb(self.raw_kdb_full_path, self.password)

        def _kdb_group_get_or_create(path):
            raise EntryBuildError(
                f"Fail while processing group {path} result: {None}"
            )

        self.processor._kdb_group_get_or_create = _kdb_group_get_or_create
        kdb_io = io.BytesIO()
        self.processor.build_kdb(kdb_io, password="123")
        kdb_io.seek(0)
        kdb = PyKeePass(kdb_io, password="123")

        self.assertEqual(len(kdb.entries), 0)
        self.assertEqual(len(kdb.groups), 1)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


def read_req(req_file):
    with open(req_file) as req:
        return [line.strip() for line in req.readlines() if line.strip() and not line.strip().startswith('#')]


requirements = read_req('requirements.txt')


setuptools.setup(
    name="kdb-vault-tools",
    version="0.1.6",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=requirements,
    author="Max Nikitenko",
    author_email="moaddib666@gmail.com",
    description="Tools for kdb vault managing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/moaddib666/kdb-vault-tools",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from unittest import TestCase

from kdb_vault_tools.datatypes.basic import BaseEntry, BaseGroup

from .utils import create_kdb, create_kdb_entry, create_kdb_group


class TestBaseGroup(TestCase):
    def setUp(self):
        self.entry = BaseGroup(title="test", path="/", icon="1")
        self.kdb = create_kdb()

    def test_creation(self):

        group = self.entry
        self.assertEqual(group.path, "/")
        self.assertEqual(group.title, "test")
        self.assertEqual(group.icon, "1")

    def test_to_dict(self):
        dict_group = self.entry.as_dict()

        self.assertIsInstance(dict_group, dict)
        self.assertEqual(
            dict_group, {"path": "/", "title": "test", "icon": "1"}
        )

    def _test_entry(self, group_name, icon, expected_path, parent_group=None):
        kdb_group = create_kdb_group(
            self.kdb,
            group_name=group_name,
            icon=icon,
            parent_group=parent_group,
        )
        group = BaseGroup.from_kdb_entry(kdb_group)
        dict_group = group.as_dict()
        self.assertEqual(
            dict_group,
            {"path": expected_path, "title": group_name, "icon": icon},
        )
        return kdb_group

    def test_from_kdb_entry(self):
        group = self._test_entry("test", "1", "test")
        group2 = self._test_entry("1", "2", "test/1", parent_group=group)
        self._test_entry("sub_sub", "3", "test/1/sub_sub", parent_group=group2)


class TestBaseEntry(TestCase):
    def setUp(self):
        self.entry = BaseEntry(
            title="test",
            path="/",
            icon="1",
            password="test123",
            url="https://127.0.0.1/",
            meta={"key": "value"},
            username="tester",
        )
        self.kdb = create_kdb()

    def test_creation(self):

        entry = self.entry
        self.assertEqual(entry.path, "/")
        self.assertEqual(entry.title, "test")
        self.assertEqual(entry.icon, "1")
        self.assertEqual(entry.password, "test123")
        self.assertEqual(entry.url, "https://127.0.0.1/")
        self.assertEqual(entry.username, "tester")
        self.assertEqual(entry.meta, {"key": "value"})

    def test_to_dict(self):
        entry = self.entry.as_dict()

        self.assertIsInstance(entry, dict)
        self.assertEqual(
            entry,
            {
                "path": "/",
                "title": "test",
                "icon": "1",
                "password": "test123",
                "url": "https://127.0.0.1/",
                "username": "tester",
                "meta": {"key": "value"},
            },
        )

    def _test_entry(
        self,
        entry_name,
        icon,
        expected_path,
        parent_group=None,
        group_name=None,
    ):

        if group_name:
            kdb_group = create_kdb_group(
                self.kdb,
                group_name=group_name,
                icon=icon,
                parent_group=parent_group or self.kdb.root_group,
            )
        elif parent_group:
            kdb_group = parent_group
        else:
            kdb_group = self.kdb.root_group

        entry_data = {
            "title": entry_name,
            "icon": icon,
            "password": "test123",
            "url": "https://127.0.0.1/",
            "username": "tester",
            "notes": "FOO: bar\n Baz Fuz",
        }
        kdb_entry = create_kdb_entry(
            self.kdb, parent_group=kdb_group, **entry_data
        )
        base_entry = BaseEntry.from_kdb_entry(kdb_entry)
        dict_entry = base_entry.as_dict()
        self.assertEqual(
            dict_entry,
            {
                "path": expected_path,
                "title": entry_name,
                "icon": icon,
                "password": "test123",
                "url": "https://127.0.0.1/",
                "username": "tester",
                "meta": {"FOO": "bar", "": "BazFuz"},
            },
        )
        return kdb_group, kdb_entry

    def test_from_kdb_entry(self):
        self._test_entry("test1", "1", "test1")
        group, _ = self._test_entry(
            "test2", "1", "foo/test2", group_name="foo"
        )
        _, entry = self._test_entry(
            "test3", "1", "foo/bar/test3", parent_group=group, group_name="bar"
        )

        base_entry = BaseEntry.from_kdb_entry(entry)
        self.assertEqual(base_entry.group_path, "foo/bar")

    def test_json_dump(self):
        entry = BaseEntry(
            title="test",
            password="123",
            url="https://1.tld",
            username="",
            path="",
            icon=None,
            meta={},
        )
        json_payload = entry.dump_json(indent=3)
        self.assertEqual(
            json_payload,
            '{\n   "title": "test",\n   "path": "",\n   "password": "123",\n   '
            '"meta": {},\n   "icon": null,\n   "username": "",\n   "url": "https://1.tld"\n}',
        )

    def test_parse_notes(self):
        notes = "test: 123"
        parsed = BaseEntry.parse_notes(notes)

        self.assertEqual(parsed, {"test": "123"})

        notes = "test: 123\nSome data here\n"
        parsed = BaseEntry.parse_notes(notes)
        self.assertEqual(parsed, {"test": "123", "Some": "datahere", "": ""})

        notes = ""
        parsed = BaseEntry.parse_notes(notes)
        self.assertEqual(parsed, {})

    def test_dump_notes(self):
        meta = {"Foo": "Bar", "Baz": {"Nested": "Not Supported yet"}}
        notes = BaseEntry.dump_notes(meta)
        self.assertEqual(
            notes, "Foo: Bar\nBaz: {'Nested': 'Not Supported yet'}"
        )



import io
import logging
import os
import sys
from copy import copy
from typing import List, Type, Union

import hvac
from hvac import exceptions as vault_errors
from pykeepass import PyKeePass, create_database
from pykeepass.entry import Entry
from pykeepass.group import Group

from kdb_vault_tools.datatypes.basic import BaseEntry, BaseGroup
from kdb_vault_tools.exceptions import EntryBuildError, VaultConnectionError

logger = logging.getLogger(__package__)
logger.level = logging.DEBUG

# TODO Configure logger
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
handler.setFormatter(formatter)
logger.addHandler(handler)


class KDBProcessor:

    kdb_entries: dict
    kdb_groups: dict

    _blank_kdb_file = "blank_database.kdbx"
    _blank_kdb_file_password = "password"
    __tmd_kdb: PyKeePass

    def exception_handler(self, exception: Exception):
        pass

    def __init__(self, *args, **kwargs):
        self.kdb_entries = {}
        self.kdb_groups = {}

    def init_kdb_connection(self):
        pass

    def load_from_kdb(self, filename: Union[str, io.BytesIO], password: str):
        self.__tmd_kdb = PyKeePass(filename=filename, password=password)
        self._load_groups_from_kdb()

    def _load_group_from_kdb(self, group: Group, load_entries=True):
        group_ = BaseGroup.from_kdb_entry(group)
        self.kdb_groups[group_.path] = group_

        if group.entries and load_entries:
            self._load_entries_from_kdb(group.entries)

    def _load_groups_from_kdb(self):
        for entry in self.kdb.groups:
            self._load_group_from_kdb(entry)

    def _load_entry_from_kdb(self, entry: Entry):
        entry_ = BaseEntry.from_kdb_entry(entry)
        self.kdb_entries.setdefault(entry_.group_path, []).append(entry_)

    def _load_entries_from_kdb(self, entries: Type[List[Entry]] = None):
        entries = entries or self.kdb.entries
        for entry in entries:
            self._load_entry_from_kdb(entry)

    def build_kdb(self, filename: Union[str, io.BytesIO], password: str):
        self._build_from_blank(filename, password)
        self._save_groups_to_kdb()
        self._save_entries_to_kdb()
        if hasattr(self.kdb.filename, "seek"):
            self.kdb.filename.seek(0)
        self.kdb.save()

    def _build_from_blank(self, filename, password):
        if self._blank_kdb_file and os.path.isfile(self._blank_kdb_file):
            kdb = PyKeePass(
                self._blank_kdb_file, self._blank_kdb_file_password
            )
            kdb.filename = filename
            kdb.password = password
            kdb.keyfile = None
        else:
            kdb = create_database(filename, password)

        self.__tmd_kdb = kdb

    def _save_groups_to_kdb(self):
        for name in self.kdb_groups.keys():
            try:
                self._kdb_group_get_or_create(name)
            except EntryBuildError as err:
                logger.exception(f"Group was not created {err}")
                continue

    def _save_entries_to_kdb(self):
        for entry_path, entry_list in self.kdb_entries.items():
            try:
                kdb_group = self._kdb_group_get_or_create(entry_path)
            except EntryBuildError as err:
                logger.exception(f"Group was not created {err}")
                continue

            for entry_ in entry_list:
                logger.debug(
                    f"Adding entry {entry_.title} to group {entry_path}"
                )
                try:
                    self.kdb.add_entry(
                        kdb_group,
                        title=entry_.title,
                        username=entry_.username or "",
                        notes=entry_.dump_notes(entry_.meta),
                        icon=entry_.icon,
                        password=entry_.password,
                        url=entry_.url or "",
                    )
                except Exception as err:  # pylint: disable=W0703
                    logger.error(f"Error while adding entry {entry_} {err}")
                    self.exception_handler(exception=err)

    def _kdb_group_get_or_create(self, path: str) -> Group:
        parent_group = self.kdb.root_group

        if path != "":
            _path = path.split("/")
            for path_entry_index, _ in enumerate(_path):
                c_path = _path[: path_entry_index + 1]
                try:
                    group = self.kdb.find_groups(path=c_path, first=True)
                except Exception as err:
                    self.exception_handler(err)
                    raise EntryBuildError(
                        f"Fail while processing group {c_path} result: {group}"
                    ) from err

                logger.debug(f"Find group {c_path} result: {group}")
                if not group:
                    j_c_path = "/".join(c_path)
                    l_group = self.kdb_groups.get(j_c_path)
                    icon = l_group.icon if l_group else None
                    group = self.kdb.add_group(
                        parent_group,
                        group_name=_path[path_entry_index],
                        icon=icon,
                    )
                parent_group = group
        return parent_group

    @property
    def kdb(self) -> PyKeePass:
        return self.__tmd_kdb


class VaultProcessor:

    kdb_entries: dict
    kdb_groups: dict

    __tmd_vault: hvac.Client
    __g_meta_label: str = "__META"

    def __init__(self, *args, **kwargs):
        self.kdb_entries = {}
        self.kdb_groups = {}

    def vault_connect(self, **kwargs):

        self._setup_connection(**kwargs)

        if not self.vault.is_authenticated():
            raise VaultConnectionError("Vault not Configured")

    def _setup_connection(self, **kwargs):
        ldap_config = kwargs.pop("ldap_config", None)

        self.__tmd_vault = hvac.Client(**kwargs)

        if ldap_config:
            self.__tmd_vault.ldap.login(**ldap_config)

    def read_secret(self, path: str):
        try:
            resp_ = self.vault.secrets.kv.v2.read_secret_version(path=path)
            logger.debug(f"Processing read secret {path}")
        except vault_errors.Forbidden as err:
            logger.debug(f"Failed to read secret {path} - {err}")
            return
        data = resp_.get("data", {}).get("data")
        path_ = copy(path).lstrip("/")
        g_path, title = path_.rsplit("/", 1) if "/" in path_ else ["", path_]

        if title == self.__g_meta_label:
            _, title = g_path.rsplit("/", 1) if "/" in g_path else ["", g_path]
            group = BaseGroup(title=title, path=g_path, **data)
            self.kdb_groups[g_path] = group
        else:
            try:
                entry = BaseEntry(**data, path=path_, title=title)
                self.kdb_entries.setdefault(g_path, []).append(
                    entry
                )  # TODO make method move to mixin
            except TypeError as err:
                logger.info(
                    f"Secret configured improperly {err} {data} {path_} {title}"
                )

    def write_secret(self, entry: BaseEntry, base_path="/"):
        data = entry.as_dict()
        data.pop("title", None)
        path = base_path + data.pop("path")
        path = path.encode("utf-8").decode("utf-8", "ignore").strip()
        try:
            self.vault.secrets.kv.create_or_update_secret(
                path=path,
                secret=data,
            )
        except (vault_errors.InvalidRequest, vault_errors.InvalidPath) as err:
            print("ERROR FUCK ", err)
            log_data = copy(data)
            log_data["password"] = "******"
            logger.exception(
                f"Cant write secret '{err}', path: '{path}' data: '{log_data}''"
            )

    def write_group(self, group: BaseGroup, base_path="/"):
        data = group.as_dict()
        data.pop("title")
        path = base_path + data.pop("path") + "/" + self.__g_meta_label

        self.vault.secrets.kv.create_or_update_secret(
            path=path,
            secret=data,
        )

    def read_secrets(self, path="/", recursive=True):

        if not path.endswith("/"):
            self.read_secret(path)
            return

        try:
            resp_ = self.vault.secrets.kv.v2.list_secrets(path=path)
        except vault_errors.InvalidPath:
            logger.info(f"Trying access invalid vault path {path}")
            return

        data = resp_.get("data")
        keys = data.get("keys")
        if recursive and keys:
            for k_name in keys:
                self.read_secrets(path=path + k_name)

    def build_vault(self, base_path="/"):
        for name in self.kdb_groups.values():
            self.write_group(name, base_path=base_path)

        for key_list in self.kdb_entries.values():
            for name in key_list:
                self.write_secret(name, base_path=base_path)

    @property
    def vault(self) -> hvac.Client:
        return self.__tmd_vault

    @property
    def group_meta_label(self) -> str:
        return self.__g_meta_label


class Processor:

    kdb_entries: dict
    kdb_groups: dict

    _kdb_processor = KDBProcessor()
    _vault_processor = VaultProcessor()

    _kdb_settings: dict
    _vault_settings: dict

    def __init__(self, vault_settings: dict, kdb_settings: dict):
        self.kdb_entries = {}
        self.kdb_groups = {}
        self._vault_processor.kdb_entries = (
            self._kdb_processor.kdb_entries
        ) = self.kdb_entries
        self._vault_processor.kdb_groups = (
            self._kdb_processor.kdb_groups
        ) = self.kdb_groups
        self._vault_settings = vault_settings
        self._kdb_settings = kdb_settings

    def sync_from_vault(self):
        self._vault_processor.vault_connect(**self._vault_settings)
        self._vault_processor.read_secrets("/")

    def write_vault(self, base_path="/"):
        self._vault_processor.vault_connect(**self._vault_settings)
        self._vault_processor.build_vault(base_path=base_path)

    def write_kdb(self):
        self._kdb_processor.build_kdb(**self._kdb_settings)

    def sync_from_kdb(self):
        self._kdb_processor.load_from_kdb(**self._kdb_settings)
